import os
from multiprocessing import Pool, cpu_count

from apiscout.ApiScout import ApiScout


class FingerprintCrawler():

    def __init__(self, db_paths, winapi1024_path, root_path, filter_size=0):
        self._scout = self._initScout(db_paths, winapi1024_path)
        self._filter_size = filter_size
        self._root_path = root_path
        self._ignore_files = {}
        self._empty_fingerprint = self.getEmptyFingerprint()
        
    def _initScout(self, db_paths, winapi1024_path):
        scout = ApiScout()
        # override potential ASLR offsets that are stored in the API DB files.
        scout.ignoreAslrOffsets(True)
        # load DB file
        for db_path in db_paths:
            scout.loadDbFile(db_path)
        # load WinApi1024 vector
        scout.loadWinApi1024(winapi1024_path)
        return scout
        
    def _getFamilyAndPath(self, path_to_binary):
        sub_path = path_to_binary[len(self._root_path):]
        path_parts = [part for part in sub_path.split(os.sep) if part]
        family = "-"
        path = ""
        if len(path_parts) > 1:
            family = path_parts[0]
            path = os.sep.join(path_parts[1:])
        else:
            path = path_parts[0]
        return family, path
        
    def getEmptyFingerprint(self):
        results = self._scout.crawl(b"\x00" * 0x100)
        filtered_results = self._scout.filter(results, 0, 0, self._filter_size)
        api_vectors = self._scout.getWinApi1024Vectors(filtered_results)
        apiscout_db_name, primary_vector = self._scout.getPrimaryVector(api_vectors)
        return primary_vector["vector"]
        
    def getFingerprint(self, path_to_binary):
        family, path = self._getFamilyAndPath(path_to_binary)
        if self._isIgnorableBinary(family, path):
            return self._ignore_files[family + "_" + path]
        if os.path.isfile(path_to_binary):
            with open(path_to_binary, "rb") as f_binary:
                binary = f_binary.read()
        # scout the binary
        try:
            results = self._scout.crawl(binary)
            filtered_results = self._scout.filter(results, 0, 0, self._filter_size)
            api_vectors = self._scout.getWinApi1024Vectors(filtered_results)
            apiscout_db_name, primary_vector = self._scout.getPrimaryVector(api_vectors)
            return (family, path, primary_vector["in_api_vector"], primary_vector["num_unique_apis"], primary_vector["vector"])
        except:
            return (family, path, 0, 0, self._empty_fingerprint)
        
    def loadIgnoreList(self, path_to_list):
        self._ignore_files = {}
        if os.path.isfile(path_to_list):
            with open(path_to_list, "r") as infile:
                for line in infile.readlines():
                    line = line.strip()
                    if line:
                        family = line.split(";")[0]
                        binary_path = line.split(";")[1]
                        self._ignore_files[family + "_" + binary_path] = tuple(line.split(";"))
                        
    def getNumberOfCrawlableFiles(self, file_list):
        return len(file_list) - self.getNumberOfIgnorableFiles(file_list)
        
    def getNumberOfIgnorableFiles(self, file_list):
        num_ignorable = 0
        for path_to_binary in file_list:
            family, path = self._getFamilyAndPath(path_to_binary)
            if self._isIgnorableBinary(family, path):
                num_ignorable += 1
        return num_ignorable
    
    def _isIgnorableBinary(self, family, path):
        return family + "_" + path in self._ignore_files
        
    def persistFingerprintsToFile(self, fingerprint_list, dest_path):
        with open(dest_path, "w") as f_destination:
            for entry in fingerprint_list:
                f_destination.write("%s;%s;%s;%s;%s\n" % entry)




########################################################################
# Copyright (c) 2020
# Daniel Plohmann <daniel.plohmann<at>mailbox<dot>org>
# All rights reserved.
########################################################################
#
#  This file is part of apiscout
#
#  apiscout is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see
#  <http://www.gnu.org/licenses/>.
#
########################################################################

import logging

LOG = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG, format="%(asctime)-15s %(message)s")

try:
    from idaapi import *
    import idc
    import idautils
    import idaapi
    if idaapi.IDA_SDK_VERSION >= 700:
        import ida_bytes
except:
    print("could not import IDA python packages - probably being used externally")
    raise



class IdaProxy(object):

    def getByte(self, ea):
        if idaapi.IDA_SDK_VERSION < 700:
            return idc.Byte(ea)
        else:
            return idc.get_wide_byte(ea)

    def getSegEnd(self, ea):
        if idaapi.IDA_SDK_VERSION < 700:
            return idc.SegEnd(ea)
        else:
            return idc.get_segm_end(ea)

    def MakeDWord(self, ea):
        if idaapi.IDA_SDK_VERSION < 700:
            return idc.MakeDword(ea)
        else:
            return ida_bytes.create_data(ea, FF_DWORD, 4, idaapi.BADADDR)

    def MakeQWord(self, ea):
        if idaapi.IDA_SDK_VERSION < 700:
            return idc.MakeQword(ea)
        else:
            return ida_bytes.create_data(ea, FF_QWORD, 8, idaapi.BADADDR)

    def MakeName(self, ea, name):
        if idaapi.IDA_SDK_VERSION < 700:
            return idc.MakeNameEx(ea, name, 256)
        else:
            return idc.set_name(ea, name, 256)

    def addTil(self, lib_name):
        if idaapi.IDA_SDK_VERSION < 700:
            return add_til(lib_name)
        else:
            return add_til(lib_name, idaapi.ADDTIL_DEFAULT)



########################################################################
# Copyright (c) 2017
# Daniel Plohmann <daniel.plohmann<at>mailbox<dot>org>
# All rights reserved.
########################################################################
#
#  This file is part of apiscout
#
#  apiscout is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see
#  <http://www.gnu.org/licenses/>.
#
########################################################################

import logging
import operator
import os
import re
import sys

from apiscout.IdaForm import IdaApiScoutOptionsForm, IdaApiScoutResultsForm
from apiscout.IdaProxy import IdaProxy

LOG = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG, format="%(asctime)-15s %(message)s")

try:
    from idaapi import *
    import idc
    import idautils
    import idaapi
    if idaapi.IDA_SDK_VERSION < 740:
        try:
            from idaapi import set_op_tinfo2 as set_op_tinfo
        except:
            LOG.error("could not import set_op_tinfo(), application of type info will not work.")
    else:
        import ida_bytes
        from ida_nalt import set_op_tinfo
except:
    LOG.error("could not import IDA python packages - probably being used externally")
    raise

def lrange(num1, num2=None, step=1):
    """
    Allows iteration over arbitrary numbers instead of dword long numbers.
    Credits go to:
    http://stackoverflow.com/questions/2187135/range-and-xrange-for-13-digit-numbers-in-python
    http://stackoverflow.com/users/263162/ricardo-cardenes
    """
    op = operator.__lt__

    if num2 is None:
        num1, num2 = 0, num1
    if num2 < num1:
        if step > 0:
            num1 = num2
        op = operator.__gt__
    elif step < 0:
        num1 = num2

    while op(num1, num2):
        yield num1
        num1 += step


class IdaTools(object):

    def init(self):
        self.ida_proxy = IdaProxy()

    def getAllMemoryFromIda(self):
        self.ida_proxy = IdaProxy()
        result = {}
        seg_start = [ea for ea in idautils.Segments()][0]
        current_start = seg_start
        seg_end = self.ida_proxy.getSegEnd(current_start)
        current_buffer = ""
        for index, current_start in enumerate(idautils.Segments()):
            # get current buffer content
            current_buffer = ""
            for ea in lrange(current_start, self.ida_proxy.getSegEnd(current_start)):
                current_buffer += chr(self.ida_proxy.getByte(ea))
            # first buffer is only saved
            if index == 0:
                result[seg_start] = current_buffer
                continue
            # otherwise decide if the buffers are consecutive and either save or update contents
            if current_start != seg_end:
                seg_start = current_start
                result[seg_start] = current_buffer
            else:
                result[seg_start] += current_buffer
            seg_end = self.ida_proxy.getSegEnd(current_start)
        # convert to bytes
        if sys.version_info > (3,):
            for segment_offset, data in result.items():
                if isinstance(data, str):
                    result[segment_offset] = bytes([ord(c) for c in data])
        return result

    def getBaseAddress(self):
        return [ea for ea in idautils.Segments()][0]

    def getLastAddress(self):
        return self.ida_proxy.getSegEnd([ea for ea in idautils.Segments()][-1]) - 1

    def makeDQWord(self, api):
        match = re.search(r"\((?P<bitness>..)bit\)", api[2])
        if match:
            bitness = int(match.group("bitness"))
        if bitness == 32:
            self.ida_proxy.MakeDWord(api[0])
        elif bitness == 64:
            self.ida_proxy.MakeQWord(api[0])

    def makeNameAndStructure(self, api, suffix=None):
        if suffix is not None:
            named = self.ida_proxy.MakeName(api[0], str(api[3] + "_{}".format(suffix)))
        else:
            named = self.ida_proxy.MakeName(api[0], str(api[3]))
        self.makeDQWord(api)
        return named

    def importTypeLibraries(self):
        if self.ida_proxy.addTil("wdk8_um") != 1 or self.ida_proxy.addTil("mssdk_win7") != 1:
            return False
        return True

    def helper_getTinfoOfFuncName(self, funcName):
        try:
            sym = til_symbol_t()
            sym.til = cvar.idati
            sym.name = funcName
            tinfo = idaapi.tinfo_t()
            namedType = get_named_type(sym.til, sym.name, 0)
            if namedType == None:
                return tinfo, False
            tinfo.deserialize(sym.til, namedType[1], namedType[2])
            return tinfo, True
        except:
            return None, False

    def setFunctionInformation(self, funcName, callAddress):
        tinfo, success = self.helper_getTinfoOfFuncName(funcName)
        if not success:
            print("Error: Cannot resolve function %s - maybe the correct type library is not yet imported?" % (funcName))
            return False
        errorCode = apply_callee_tinfo(callAddress, tinfo) #in IDA 6.9 this returns <type 'NoneType'>, in IDA 7.1 it is "True"
        try:
            success = set_op_tinfo(callAddress, 0, tinfo)
        except:
            LOG.error("Could not set type info, set_op_tinfo() not available.")
        if errorCode not in [None, True] or not success:
            return False
        return True

    def applyApiNames(self, api_results):
        num_renamed = 0
        num_skipped = 0
        num_xrefs_adapted = 0

        prev_offset = 0
        for api in sorted(api_results):
            if api[0] > prev_offset + 16:
                print("Annotating API Block @0x{:x}.".format(api[0]))
            prev_offset = api[0]
            if str(api[3]) == "None":
                num_skipped += 1
                print("Skipping 0x{:x}: no name provided by API DB (is None).".format(api[0]))
                self.makeDQWord(api)
                continue
            named = self.makeNameAndStructure(api)
            if not named:
                for suffix in range(10):
                    print("naming 0x{:x} to {} failed, trying with suffix \"_{}\".".format(api[0], str(api[3]), suffix))
                    named = self.makeNameAndStructure(api, suffix)
                    if named:
                        break
                    else:
                        print("  naming 0x{:x} to {} failed as well, trying next index...".format(api[0], str(api[3] + "_{}".format(suffix))))
            if named:
                num_renamed += 1
                for xref in idautils.XrefsTo(api[0]):
                    if self.setFunctionInformation(api[3], xref.frm):
                        num_xrefs_adapted += 1
        return num_renamed, num_skipped, num_xrefs_adapted

    def formGetParameters(self):
        parameters = {}
        this_dir = os.path.abspath(os.path.join(os.path.dirname(__file__)))
        db_folder = os.path.abspath(os.path.join(this_dir, "..", "dbs"))

        form = IdaApiScoutOptionsForm(db_folder)
        form.Compile()
        ok = form.Execute()
        if ok == 1:
            if form.chosenValues:
                parameters["api_dbs"] = form.chosenValues
            else:
                parameters["api_dbs"] = [form.iFileOpen.value]
            parameters["ignore_aslr_offset"] = form.rAslr.checked
        form.Free()
        return parameters

    def formSelectResults(self, results):
        selected_apis = []

        form = IdaApiScoutResultsForm(results, self.getBaseAddress(), self.getLastAddress())
        form.Compile()
        ok = form.Execute()
        if ok == 1 and form.chosenApis:
            selected_apis = form.chosenApis
        form.Free()
        return selected_apis




# -----------------------------------------------------------------------
# based on:
# https://github.com/idapython/src/blob/master/examples/ex_askusingform.py
# (c) Hex-Rays
#
import json
import os

import idautils
import idaapi
if idaapi.IDA_SDK_VERSION < 700:
    from idaapi import Form
    from idaapi import Choose2 as Choose
else:
    from idaapi import Form, Choose

from apiscout.ApiScout import ApiScout


class ApiDbChooser(Choose):
    """
    A simple chooser to be used as an embedded chooser
    """
    def __init__(self, title, api_db_folder, flags=0):
        Choose.__init__(self,
                         title,
                         [["Filename", 25], ["OS Version", 20], ["ASLR Offset?", 8], ["DLLs", 6], ["APIs", 6]],
                         embedded=True, width=120, height=10, flags=flags)
        self.row_count = 0
        self.api_db_folder = api_db_folder
        self.items = self.populate(api_db_folder)
        self.icon = 4
        self.selcount = 0

    def populate(self, api_db_folder):
        api_dbs = []
        for filename in os.listdir(api_db_folder):
            if not filename.endswith(".json"):
                continue
            db_data = {}
            with open(api_db_folder + os.sep + filename, "r") as f_api_file:
                db_data = json.loads(f_api_file.read())
            api_dbs.append([filename, db_data["os_version"], "%s" % db_data["aslr_offsets"], "%d" % db_data["num_dlls"], "%d" % db_data["num_apis"]])
            self.row_count += 1
        return api_dbs

    def OnClose(self):
        pass

    def getItems(self, indices):
        items = []
        for index in indices:
            items.append(self.api_db_folder + os.sep + self.items[index][0])
        return items

    def OnGetLine(self, n):
        return self.items[n]

    def OnGetSize(self):
        n = len(self.items)
        return n

class IdaApiScoutOptionsForm(Form):

    def __init__(self, api_db_folder):
        self.invert = False
        self.chosenValues = []
        self.apiDbChooser = ApiDbChooser("ApiDBs", api_db_folder, flags=Choose.CH_MULTI)
        Form.__init__(self, r"""STARTITEM {id:rNormal}
BUTTON YES* Run
BUTTON CANCEL Cancel
IDA ApiScout

{FormChangeCb}
Please select one or more API DBs from your apiscout/dbs folder:

<Available API DBs:{cApiDbChooser}>

or load a database from another location:

<#Select a file to open#:{iFileOpen}>

<##Ignore ASLR offsets:{rAslr}>{cGroup1}>
""", {
    'iAslrOffset': Form.NumericInput(tp=Form.FT_UINT64, value=0x0),
    'iFileOpen': Form.FileInput(swidth=40, open=True, value="*.*"),
    'cGroup1': Form.ChkGroupControl(("rAslr", "rNormal")),
    'FormChangeCb': Form.FormChangeCb(self.OnFormChange),
    'cApiDbChooser' : Form.EmbeddedChooserControl(self.apiDbChooser)
})

    def OnFormChange(self, fid):
        if fid == self.cApiDbChooser.id:
            indices = self.GetControlValue(self.cApiDbChooser)
            self.chosenValues = self.apiDbChooser.getItems(indices)
        return 1


class ApiChooser(Choose):
    """
    A simple chooser to be used as an embedded chooser
    """
    def __init__(self, title, api_results, flags=0):
        Choose.__init__(self,
                         title,
                         [["#", 6], ["Offset", 14], ["API Address", 14], ["DLL", 20], ["API", 35]],
                         embedded=True, width=140, height=20, flags=flags)
        self.row_count = 0
        self.base_address = [ea for ea in idautils.Segments()][0]
        self.scout = ApiScout()
        self.scout.setBaseAddress(self.base_address)
        self.api_results = api_results
        self.all_items = self.populate(api_results)
        self.items = self.populate(api_results)
        self.icon = 4
        self.selcount = 0

    def filterDisplay(self, from_addr, to_addr, distance):
        filtered_items = self.scout.filter(self.api_results, from_addr, to_addr, distance)
        self.items = self.populate(filtered_items)

    def populate(self, api_results):
        api_rows = []
        unified_results = set([])
        for key in api_results:
            unified_results.update(api_results[key])
        for index, entry in enumerate(sorted(unified_results)):
            dll_name = "{} ({}bit)".format(entry[2], entry[4])
            api_rows.append(["%d" % (index + 1), "0x%x" % (self.base_address + entry[0]), "0x%x" % entry[1], dll_name, entry[3]])
            self.row_count += 1
        return api_rows

    def OnClose(self):
        pass

    def getItems(self, l):
        items = []
        for index in l:
            items.append([int(self.items[index][1], 16), int(self.items[index][2], 16), self.items[index][3], str(self.items[index][4])])
        return items

    def OnGetLine(self, n):
        return self.items[n]

    def OnGetSize(self):
        n = len(self.items)
        return n


class IdaApiScoutResultsForm(Form):

    def __init__(self, crawled_apis, from_addr=0, to_addr=0):
        self.invert = False
        self.chosenApis = []
        self.apiChooser = ApiChooser("Apis", crawled_apis, flags=Choose.CH_MULTI)
        Form.__init__(self, r"""STARTITEM {id:rNormal}
BUTTON YES* Annotate
BUTTON CANCEL Cancel
IDA ApiScout (Results)

{FormChangeCb}
ApiScout has found the following APIs (select to annotate, e.g. CTRL+A):

<APIs:{cApiChooser}>
{cApiInfo}

Filter APIs by Range -  
<##from  :{iAddrFrom}>
<##to    :{iAddrTo}>
Filter APIs by Grouping - require another API 
<##within:{iByteRange}> bytes
<##Apply Filter:{bApplyFilter}>
""", {
    'cApiInfo': Form.StringLabel("APIs"),
    'bApplyFilter': Form.ButtonInput(self.OnButtonApplyFilter),
    'iAddrFrom': Form.NumericInput(tp=Form.FT_ADDR, value=from_addr),
    'iAddrTo': Form.NumericInput(tp=Form.FT_ADDR, value=to_addr),
    'iByteRange': Form.NumericInput(tp=Form.FT_UINT64, value=0x800),
    'cGroup1': Form.ChkGroupControl(("rFilter", "rNormal")),
    'FormChangeCb': Form.FormChangeCb(self.OnFormChange),
    'cApiChooser' : Form.EmbeddedChooserControl(self.apiChooser)
})

    def OnButtonApplyFilter(self, code=0):
        addr_from = self.GetControlValue(self.iAddrFrom)
        addr_to = self.GetControlValue(self.iAddrTo)
        byte_range = self.GetControlValue(self.iByteRange)
        self.apiChooser.filterDisplay(addr_from, addr_to, byte_range)
        # SetControlValue seems to have broken in some API migration...
        # if addr_from == 0 and addr_to == 0 and byte_range == 0:
        #     self.SetControlValue(self.cApiInfo, "APIs: %d/%d (unfiltered)" % (len(self.apiChooser.items), len(self.apiChooser.all_items)))
        # else:
        #     self.SetControlValue(self.cApiInfo, "APIs: %d/%d (filtered to 0x%x - 0x%x, range: 0x%x)" % (len(self.apiChooser.items), len(self.apiChooser.all_items), addr_from, addr_to, byte_range))
        l = self.GetControlValue(self.cApiChooser)
        self.chosenApis = self.apiChooser.getItems(l)
        self.RefreshField(self.cApiChooser)

    def OnFormChange(self, fid):
        if fid == self.cApiChooser.id:
            l = self.GetControlValue(self.cApiChooser)
            self.chosenApis = self.apiChooser.getItems(l)
        return 1

    def OnClose(self):
        l = self.GetControlValue(self.cApiChooser)
        self.chosenApis = self.apiChooser.getItems(l)



import logging
import struct

from .utility import get_word

if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOG = logging.getLogger(__name__)

class PeTools(object):

    BITNESS_MAP = {0x14c: 32, 0x8664: 64}

    @staticmethod
    def mapBinary(binary):
        # This is a pretty rough implementation but does the job for now
        mapped_binary = bytearray([])
        pe_offset = PeTools.getPeOffset(binary)
        if pe_offset:
            num_sections = 0
            bitness = 0
            section_infos = []
            optional_header_size = 0xF8
            if pe_offset and len(binary) >= pe_offset + 0x8:
                num_sections = struct.unpack("H", binary[pe_offset + 0x6:pe_offset + 0x8])[0]
                bitness = PeTools.getBitness(binary)
                if bitness == 64:
                    optional_header_size = 0x108
            if pe_offset and num_sections and len(binary) >= pe_offset + optional_header_size + num_sections * 0x28:
                for section_index in range(num_sections):
                    section_offset = section_index * 0x28
                    slice_start = pe_offset + optional_header_size + section_offset + 0x8
                    slice_end = pe_offset + optional_header_size + section_offset + 0x8 + 0x10
                    virt_size, virt_offset, raw_size, raw_offset = struct.unpack("IIII", binary[slice_start:slice_end])
                    section_info = {
                        "section_index": section_index,
                        "virt_size": virt_size,
                        "virt_offset": virt_offset,
                        "raw_size": raw_size,
                        "raw_offset": raw_offset,
                    }
                    section_infos.append(section_info)
            max_virt_section_offset = 0
            min_raw_section_offset = 0xFFFFFFFF
            if section_infos:
                for section_info in section_infos:
                    max_virt_section_offset = max(max_virt_section_offset, section_info["virt_size"] + section_info["virt_offset"])
                    max_virt_section_offset = max(max_virt_section_offset, section_info["raw_size"] + section_info["virt_offset"])
                    if section_info["raw_offset"] > 0x200:
                        min_raw_section_offset = min(min_raw_section_offset, section_info["raw_offset"])
            if max_virt_section_offset:
                mapped_binary = bytearray([0] * max_virt_section_offset)
                mapped_binary[0:min_raw_section_offset] = binary[0:min_raw_section_offset]
            for section_info in section_infos:
                mapped_binary[section_info["virt_offset"]:section_info["virt_offset"] + section_info["raw_size"]] = binary[section_info["raw_offset"]:section_info["raw_offset"] + section_info["raw_size"]]
                LOG.debug("Mapping %d: raw 0x%x (0x%x bytes) -> virtual 0x%x (0x%x bytes)", section_info["section_index"], section_info["raw_offset"], section_info["raw_size"], section_info["virt_offset"], section_info["virt_size"])
        LOG.debug("Mapped binary of size %d bytes (%d sections) to memory view of size %d bytes", len(binary), num_sections, len(mapped_binary))
        return bytes(mapped_binary)

    @staticmethod
    def getBitness(binary):
        bitness_id = 0
        pe_offset = PeTools.getPeOffset(binary)
        if pe_offset:
            if pe_offset and len(binary) >= pe_offset + 0x6:
                bitness_id = struct.unpack("H", binary[pe_offset + 0x4:pe_offset + 0x6])[0]
        return PeTools.BITNESS_MAP.get(bitness_id, 0)

    @staticmethod
    def getBaseAddressFromPeHeader(binary):
        pe_offset = PeTools.getPeOffset(binary)
        if pe_offset:
            if pe_offset and len(binary) >= pe_offset + 0x38:
                base_addr = struct.unpack("I", binary[pe_offset + 0x34:pe_offset + 0x38])[0]
                LOG.debug("Changing base address from 0 to: 0x%x for inference of reference counts (based on PE header)", base_addr)
                return base_addr
        return 0

    @staticmethod
    def getPeOffset(binary):
        if len(binary) >= 0x40:
            pe_offset = get_word(binary, 0x3c)
            return pe_offset
        return 0

    @staticmethod
    def checkPe(binary):
        pe_offset = PeTools.getPeOffset(binary)
        if pe_offset and len(binary) >= pe_offset + 6:
            bitness = get_word(binary, pe_offset + 4)
            return bitness in PeTools.BITNESS_MAP
        return False




########################################################################
# Copyright (c) 2018
# Daniel Plohmann <daniel.plohmann<at>mailbox<dot>org>
# Steffen Enders <steffen<at>enders<dot>nrw>
# All rights reserved.
########################################################################
#
#  This file is part of apiscout
#
#  apiscout is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see
#  <http://www.gnu.org/licenses/>.
#
########################################################################

import re
import struct
import os
import json
import operator
import logging

try:
    import lief
except:
    print("lief is not installed! We recommend installing lief to improve import table parsing capabilities of ApiScout!")
    lief = None

from .ImportTableLoader import ImportTableLoader
from .ApiVector import ApiVector
from .PeTools import PeTools

# Only do basicConfig if no handlers have been configured
if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOG = logging.getLogger(__name__)


class ApiScout(object):

    def __init__(self, db_filepath=None):
        self.api_maps = {}
        self.has_64bit = False
        self.base_address = 0
        self._binary_length = 0
        # Used to achieve coherent offset view in IdaScout
        self.load_offset = 0
        self.ignore_aslr_offsets = False
        self._import_table = None
        if db_filepath:
            self.loadDbFile(db_filepath)
        self._apivector = ApiVector()
        self.loadWinApi1024()

    def loadDbFile(self, db_filepath):
        api_db = {}
        if os.path.isfile(db_filepath):
            with open(db_filepath, "r") as f_json:
                api_db = json.loads(f_json.read())
        else:
            LOG.error("Not a file: %s!", db_filepath)
            raise ValueError
        num_apis_loaded = 0
        num_collisions = 0
        api_map = {}
        for dll_entry in api_db["dlls"]:
            LOG.debug("  building address map for: %s", dll_entry)
            aslr_offset = 0
            if not self.ignore_aslr_offsets:
                aslr_offset = api_db["dlls"][dll_entry]["aslr_offset"]
            for export in api_db["dlls"][dll_entry]["exports"]:
                num_apis_loaded += 1
                api_name = "%s" % (export["name"])
                if api_name == "None":
                    api_name = "None<{}>".format(export["ordinal"])
                dll_name = "_".join(dll_entry.split("_")[2:])
                bitness = api_db["dlls"][dll_entry]["bitness"]
                self.has_64bit |= bitness == 64
                base_address = api_db["dlls"][dll_entry]["base_address"]
                virtual_address = base_address + export["address"] - aslr_offset
                if virtual_address in api_map and (api_map[virtual_address][0].lower() != dll_name.lower()):
                    num_collisions += 1
                api_map[virtual_address] = (dll_name, api_name, bitness)
            LOG.debug("loaded %d exports", num_apis_loaded)
        LOG.info("loaded %d exports from %d DLLs (%s) with %d potential collisions.", num_apis_loaded, len(api_db["dlls"]), api_db["os_name"], num_collisions)
        self.api_maps[api_db["os_name"]] = api_map

    def loadWinApi1024(self, winapi1024_filepath=None):
        if winapi1024_filepath is None:
            this_dir = os.path.abspath(os.path.join(os.path.dirname(__file__)))
            winapi1024_filepath =  this_dir + os.sep + "data" + os.sep + "winapi1024v1.txt"
        self._apivector = ApiVector(winapi1024_filepath)

    def _resolveApiByAddress(self, api_map_name, absolute_addr):
        api_entry = ("", "", "")
        api_map = self.api_maps[api_map_name]
        check_address = absolute_addr
        if check_address in api_map:
            api_entry = api_map[check_address]
        return api_entry

    def getNumApisLoaded(self):
        return sum([len(self.api_maps[api_map_name]) for api_map_name in self.api_maps])

    def ignoreAslrOffsets(self, value):
        self.ignore_aslr_offsets = value

    def setBaseAddress(self, address):
        self.base_address = address

    def setLoadOffset(self, offset):
        self.load_offset = offset

    def iterateAllDwords(self, binary):
        for offset, _ in enumerate(binary):
            try:
                dword = struct.unpack("I", binary[offset:offset + 4])[0]
                yield offset, dword
            except struct.error:
                break

    def iterateAllQwords(self, binary):
        for offset, _ in enumerate(binary):
            try:
                dword = struct.unpack("Q", binary[offset:offset + 8])[0]
                yield offset, dword
            except struct.error:
                break

    def _parseImportTable(self, binary):
        if self._import_table is None:
            it_loader = ImportTableLoader(binary)
            self._import_table = it_loader.get_import_table()

    def _isImportTableEntry(self, offset):
        if not self._import_table:
            return None
        else:
            return offset in self._import_table

    def _findBaseAddress(self, binary):
        if self.base_address:
            return self.base_address
        # try to extract from PE header
        return PeTools.getBaseAddressFromPeHeader(binary)

    def _updateCodeReferences(self, references, binary, base_address, offset):
        # treat as 32bit code (absolute offsets)
        addr_block = binary[offset + 2:offset + 2 + 4]
        function_addr = struct.unpack("I", addr_block)[0] - base_address
        if 0 < function_addr < len(binary):
            if function_addr not in references[32]:
                references[32][function_addr] = 0
            references[32][function_addr] += 1
        # treat as 64bit code (this means relative offsets for jmps/calls)
        addr_block = binary[offset + 2:offset + 2 + 4]
        function_addr = struct.unpack("i", addr_block)[0]
        # we need to calculate RIP + offset + 7 (48 ff 25 ** ** ** **)
        if binary[offset:offset + 2] == "\xFF\x25":
            function_addr += offset + 7
        elif binary[offset:offset + 2] == "\xFF\x15":
            function_addr += offset + 6
        if 0 < function_addr < len(binary):
            if function_addr not in references[64]:
                references[64][function_addr] = 0
            references[64][function_addr] += 1

    def _getCodeReferences(self, binary):
        references = {32: {}, 64: {}}
        base_address = self._findBaseAddress(binary)
        # also check for "jmp dword ptr <offset>", as they sometimes point to local functions (i.e. non-API)
        for match in re.finditer(b"\xFF\x25", binary):
            self._updateCodeReferences(references, binary, base_address, match.start())
        # also check for "call dword ptr <offset>", as they sometimes point to local functions (i.e. non-API)
        for match in re.finditer(b"\xFF\x15", binary):
            self._updateCodeReferences(references, binary, base_address, match.start())
        return references

    def evaluateImportTable(self, binary, is_unmapped=True):
        self._binary_length = len(binary)
        results = {"import_table": []}
        if lief:
            lief_binary = lief.parse(bytearray(binary))
            bitness = 32 if lief_binary.header.machine == lief.PE.MACHINE_TYPES.I386 else 64
            for imported_library in lief_binary.imports:
                for func in imported_library.entries:
                    if func.name:
                        results["import_table"].append((func.iat_address + self.load_offset, 0xFFFFFFFF, imported_library.name.lower() + "_0x0", func.name, bitness, True, 1))
        else:
            # fallback using the old method and out own import table parser
            mapped_binary = binary
            if is_unmapped:
                LOG.debug("Mapping unmapped binary before processing")
                mapped_binary = PeTools.mapBinary(binary)
            bitness = PeTools.getBitness(mapped_binary)
            self._import_table = None
            self._parseImportTable(mapped_binary)
            references = self._getCodeReferences(mapped_binary)
            for offset, import_entry in sorted(self._import_table.items()):
                ref_count = 1
                if bitness:
                    ref_count = 1 + references[bitness][offset] if offset in references[bitness] else 1
                results["import_table"].append((offset + self.load_offset, 0xFFFFFFFF, import_entry["dll_name"].lower() + "_0x0", import_entry["name"], bitness, True, ref_count))
        return results

    def crawl(self, binary):
        results = {}
        self._binary_length = len(binary)
        self._import_table = None
        self._parseImportTable(binary)
        self._isImportTableEntry(0)
        references = self._getCodeReferences(binary)
        for api_map_name in self.api_maps:
            recovered_apis = []
            for offset, api_address in self.iterateAllDwords(binary):
                dll, api, bitness = self._resolveApiByAddress(api_map_name, api_address)
                if dll and api and bitness == 32:
                    ref_count = 1 + references[32][offset] if offset in references[32] else 1
                    recovered_apis.append((offset + self.load_offset, api_address, dll, api, bitness, self._isImportTableEntry(offset), ref_count))
            if self.has_64bit:
                for offset, api_address in self.iterateAllQwords(binary):
                    dll, api, bitness = self._resolveApiByAddress(api_map_name, api_address)
                    if dll and api and bitness == 64:
                        ref_count = 1 + references[64][offset] if offset in references[64] else 1
                        recovered_apis.append((offset + self.load_offset, api_address, dll, api, bitness, self._isImportTableEntry(offset), ref_count))
            results[api_map_name] = recovered_apis
        return results

    def _getFilteredList(self, distance, list_to_filter):
        offsets_a = [item[0] for item in list_to_filter]
        offsets_b = offsets_a[1:] + [0]
        api_distances = list(map(operator.sub, offsets_b, offsets_a))
        distance_filtered = []
        for index, api_distance in enumerate(api_distances[:-1]):
            if api_distance <= distance:
                if list_to_filter[index] not in distance_filtered:
                    distance_filtered.append(list_to_filter[index])
                if list_to_filter[index + 1] not in distance_filtered:
                    distance_filtered.append(list_to_filter[index + 1])
        return distance_filtered

    def filter(self, result, from_addr, to_addr, distance, own_image=True):
        filtered_result = {}
        for key in result:
            filtered_list = result[key]
            # filter all entries pointing into our own image
            if own_image:
                filtered_list = [item for item in filtered_list if (self.base_address > item[1] or (self.base_address + self._binary_length) < item[1])]
            if from_addr:
                filtered_list = [item for item in filtered_list if self.base_address + item[0] >= from_addr]
            if to_addr:
                filtered_list = [item for item in filtered_list if self.base_address + item[0] <= to_addr]
            if distance:
                if len(filtered_list) < 2:
                    filtered_list = []
                else:
                    filtered_list = self._getFilteredList(distance, filtered_list)
            filtered_result[key] = filtered_list
        return filtered_result

    def getWinApi1024Vectors(self, results):
        return self._apivector.getApiVectors(results)

    def matchVectors(self, vector_a, vector_b):
        return self._apivector.matchVectors(vector_a, vector_b)

    def matchVectorCollection(self, vector, collection_path):
        return self._apivector.matchVectorCollection(vector, collection_path)

    def getPrimaryVector(self, api_vectors):
        return sorted(api_vectors.items(), key=lambda x: x[1]["percentage"])[-1]

    def render(self, results):
        output = ""
        for api_map_name in results:
            if len(results[api_map_name]):
                result = results[api_map_name]
                output += "Results for API DB: {}\n".format(api_map_name)
                output += "{:3}: {:10}; {:18}; {:3}; {:4};{:40}; {:60}\n".format("idx", "offset", "VA", "IT?", "#ref", "DLL", "API")
                prev_offset = 0
                dlls = set()
                apis = set()
                num_references = 0
                for index, entry in enumerate(result):
                    if prev_offset and entry[0] > prev_offset + 16:
                        output += "-" * 129 + "\n"
                    dll_name = "{} ({}bit)".format(entry[2], entry[4])
                    if entry[5] is None:
                        is_in_import_table = "err"
                    else:
                        is_in_import_table = "yes" if entry[5] else "no"
                    if entry[4] == 32:
                        output += "{:3}: 0x{:08x};         0x{:08x}; {:3}; {:4}; {:40}; {:60}\n".format(index + 1, self.base_address + entry[0], entry[1], is_in_import_table, entry[6], dll_name, entry[3])
                    else:
                        output += "{:3}: 0x{:08x}; 0x{:016x}; {:3}; {:4}; {:40}; {:60}\n".format(index + 1, self.base_address + entry[0], entry[1], is_in_import_table, entry[6], dll_name, entry[3])
                    num_references += entry[6] - 1
                    prev_offset = entry[0]
                    dlls.add(entry[2])
                    apis.add(entry[3])
                output += "DLLs: {}, APIs: {}, references: {}\n".format(len(dlls), len(apis), num_references)
            else:
                output += "No results for API map: {}\n".format(api_map_name)
        return output

    def renderVectorResults(self, results):
        api_vectors = self.getWinApi1024Vectors(results)
        output = "WinApi1024 Vector Results:\n"
        for api_map_name, result in sorted(api_vectors.items()):
            output += "{}: {} / {} ({:5.2f}%) APIs covered in WinApi1024 vector.\n".format(api_map_name, result["in_api_vector"], result["num_unique_apis"], result["percentage"])
            output += "    Vector:     {}\n".format(result["vector"])
            output += "    Confidence: {}\n".format(self._apivector.getVectorConfidence(result["vector"]))
        return output

    def renderResultsVsCollection(self, results, collection_file):
        # find primary vector
        api_vectors = self.getWinApi1024Vectors(results)
        primary_vector = self.getPrimaryVector(api_vectors)
        output = "Using resulting Vector from DB \"{}\" for matching...\n".format(primary_vector[0])
        collection_result = self.matchVectorCollection(primary_vector[1]["vector"], collection_file)
        output += self.renderVectorCollectionResults(collection_result)
        return output

    def renderVectorCollectionResults(self, results, max_results=5):
        output = "WinApi1024 Vector vs Collection Results:\n"
        output += "    Vector: {}\n".format(results["vector"])
        output += "    Collection: {} ({} families, {} vectors)\n".format(results["collection_filepath"], results["families_in_collection"], results["vectors_in_collection"])
        family_width = max([len(entry[0]) for entry in results["match_results"]])
        sample_width = max([len(entry[1]) for entry in results["match_results"]])
        output += "-" * (family_width + sample_width + 5 + 6) + "\n"
        output += "Top {} family matches\n".format(max_results)
        num_results = 0
        seen_families = []
        for result in results["match_results"]:
            if num_results > max_results:
                break
            if result[0] not in seen_families:
                output += "{:{fw}} - {:{sw}} - {:.3f}\n".format(result[0], result[1], result[2], fw=family_width, sw=sample_width)
                seen_families.append(result[0])
                num_results += 1
        output += "-" * (family_width + sample_width + 5 + 6) + "\n"
        output += "Top {} individual matches\n".format(max_results)
        num_results = 0
        for result in results["match_results"]:
            if num_results > max_results:
                break
            output += "{:{fw}} - {:{sw}} - {:.3f}\n".format(result[0], result[1], result[2], fw=family_width, sw=sample_width)
            num_results += 1
        return output
